#! /usr/bin/env node
'use strict';

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _directoryTraverser = require('directory-traverser');

var _directoryTraverser2 = _interopRequireDefault(_directoryTraverser);

var _module_resolver = require('./module_resolver');

var _utils = require('./utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const srcRoot = process.cwd();
const args = process.argv;
const fromFile = (0, _utils.toAbsolute)(args[2], srcRoot);
let toFile = (0, _utils.toAbsolute)(args[3], srcRoot);

(function checkFromFile() {
  _fs2.default.statSync(fromFile);
})();

(function checkAndFixToFile() {
  try {
    const stat2 = _fs2.default.statSync(toFile);
    if (stat2.isDirectory()) {
      toFile = _path2.default.join(toFile, _path2.default.basename(fromFile));
      if (_fs2.default.existsSync(toFile)) {
        throw new Error('目标文件已存在，不允许覆盖');
      }
    } else {
      help();
      throw new Error(`toFile already exists and is a plain file: ${toFile}`);
    }
  } catch (e) {
    if (e.code === 'ENOENT') {
      return;
    }
    throw e;
  }
})();

const isDirectory = _fs2.default.statSync(fromFile).isDirectory();
const subdirOld = isDirectory && fromFile;
const subdirNew = isDirectory && toFile;

function help() {
  console.error('Usage: js-refactor <from-file-or-directory> <to-file-or-directory>');
}

const MATCH_REQUIRE = /require\s*\(\s*['"]([\s\S]+?)['"]\s*\)/g;
const MATCH_IMPORT = /import[\s\S]+?from\s*['"](.+?)['"]/g;
const MATCH_EXPORT = /export[\s\S]+?from\s*['"](.+?)['"]/g;
const IGNORE_DIR = /\/node_modules\/?$/;
const SELECT_SOURCE_FILE = /\.js$/;

// traverse source file
function iterateSourceFiles(dir, options = {}, callback) {
  const { excludes = [] } = options;
  const filter = d => {
    if (!excludes || !excludes.length) {
      return true;
    }
    return !excludes.some(matcher => {
      if (typeof matcher === 'function') {
        return matcher(d);
      } else if (matcher instanceof RegExp) {
        return matcher.test(d);
      }
      // 断言
      throw new Error('not a valid matcher');
    });
  };

  (0, _directoryTraverser2.default)(dir, filter, (subdir, filenames) => {
    filenames.forEach(f => {
      if (!SELECT_SOURCE_FILE.test(f)) {
        return;
      }
      callback({ file: _path2.default.join(subdir, f) });
    });
  });
}

function fixOnlyFile() {
  if (((0, _utils.getSuffix)(toFile) || '').toLowerCase() !== 'js') {
    return;
  }
  console.log('fix only file...\n\n');
  let fileContent = _fs2.default.readFileSync(fromFile, { encoding: 'utf8' });
  [MATCH_IMPORT, MATCH_REQUIRE, MATCH_EXPORT].forEach(re => {
    re.lastIndex = 0;
    let m;
    while ((m = re.exec(fileContent)) !== null) {
      const dep = m[1];
      const isRela = (0, _utils.isRelative)(dep);
      // 移动单个文件不影响该文件对其他module的绝对路径导入
      if (!isRela) {
        continue;
      }
      const dirOld = _path2.default.dirname(fromFile);
      const dirNew = _path2.default.dirname(toFile);
      if ((0, _utils.isSameFile)(dirOld, dirNew)) {
        return;
      }
      const realDep = _path2.default.join(dirOld, dep);
      let newDep = _path2.default.relative(dirNew, realDep);
      if (isRela && !(0, _utils.isRelative)(newDep)) {
        newDep = './' + newDep;
      }

      console.log('update file: %s\n  %s\n->%s\n', toFile, dep, newDep);
      fileContent = fileContent.substr(0, m.index) + fileContent.substr(m.index).replace(dep, newDep);
      re.lastIndex += newDep.length - dep.length;
    }
  });
  _fs2.default.writeFileSync(fromFile, fileContent);
}

function fixFilesDependingOnOnlyFile() {
  console.log('fix files depending on only file...\n\n');
  iterateSourceFiles(srcRoot, { excludes: [IGNORE_DIR] }, ({ file }) => {
    if ((0, _utils.isSameFile)(file, fromFile)) {
      return;
    }
    const dir = _path2.default.dirname(file);
    let fileContent = _fs2.default.readFileSync(file, { encoding: 'utf8' });
    [MATCH_IMPORT, MATCH_REQUIRE, MATCH_EXPORT].forEach(re => {
      re.lastIndex = 0;
      let m;
      while ((m = re.exec(fileContent)) !== null) {
        const dep = m[1];
        // 其他模块/文件/包，忽略
        const isRela = (0, _utils.isRelative)(dep);
        const resolvePath = (0, _module_resolver.resolveDep)(dir, dep);
        if (!resolvePath) {
          continue;
        }

        if (!(0, _utils.depMatchFs)(resolvePath, fromFile)) {
          continue;
        }
        const fixDep = (0, _utils.fsToDep)(toFile);
        let newDep = isRela ? _path2.default.relative(dir, fixDep) : (0, _module_resolver.useImplicitPrefix)(fixDep);
        if (isRela && !(0, _utils.isRelative)(newDep)) {
          newDep = './' + newDep;
        }
        console.log('update file: %s\n  %s\n->%s\n', file, dep, newDep);
        fileContent = fileContent.substr(0, m.index) + fileContent.substr(m.index).replace(dep, newDep);
        re.lastIndex += newDep.length - dep.length;
      }
    });
    _fs2.default.writeFileSync(file, fileContent);
  });
}

// fix files within subdir
function fixFilesInSubdir() {
  console.log('processing internal files...\n\n');
  iterateSourceFiles(subdirOld, { excludes: [IGNORE_DIR] }, ({ file }) => {
    const dir = _path2.default.dirname(file);
    const wouldBePath = _path2.default.join(subdirNew, _path2.default.relative(subdirOld, file));
    const wouldBeDir = _path2.default.dirname(wouldBePath);
    let fileContent = _fs2.default.readFileSync(file, { encoding: 'utf8' });
    [MATCH_IMPORT, MATCH_REQUIRE, MATCH_EXPORT].forEach(re => {
      re.lastIndex = 0;
      let m;
      while ((m = re.exec(fileContent)) !== null) {
        const dep = m[1];
        // 其他模块/文件/包，忽略
        const resolvePath = (0, _module_resolver.resolveDep)(dir, dep);
        if (!resolvePath) {
          continue;
        }
        const isRela = (0, _utils.isRelative)(dep);
        const isInnerFile = (0, _utils.isSuperDir)(subdirOld, resolvePath);
        if (isRela && isInnerFile || !isRela && !isInnerFile) {
          continue;
        }
        let newDep = '';
        if (isRela) {
          newDep = _path2.default.normalize(_path2.default.relative(wouldBeDir, _path2.default.join(dir, dep)));
        } else {
          newDep = (0, _module_resolver.useImplicitPrefix)((0, _utils.fsToDep)(_path2.default.join(subdirNew, _path2.default.relative(subdirOld, resolvePath))));
        }

        console.log('update file: %s\n  %s\n->%s\n', file, dep, newDep);
        fileContent = fileContent.substr(0, m.index) + fileContent.substr(m.index).replace(dep, newDep);
        re.lastIndex += newDep.length - dep.length;
      }
    });
    _fs2.default.writeFileSync(file, fileContent);
  });
}

// fix files outside subdir
function fixFilesOutsideSubdir() {
  console.log('processing external files...\n\n');
  iterateSourceFiles(srcRoot, { excludes: [d => (0, _utils.isSuperDir)(subdirOld, d), IGNORE_DIR] }, ({ file }) => {

    const dir = _path2.default.dirname(file);
    let fileContent = _fs2.default.readFileSync(file, { encoding: 'utf8' });

    [MATCH_IMPORT, MATCH_REQUIRE, MATCH_EXPORT].forEach(re => {
      re.lastIndex = 0;
      let m;
      while ((m = re.exec(fileContent)) !== null) {
        const dep = m[1];
        // 其他模块/文件/包，忽略
        const resolvePath = (0, _module_resolver.resolveDep)(dir, dep);
        if (!resolvePath) {
          continue;
        }
        const isRela = (0, _utils.isRelative)(dep);
        const isInnerFile = (0, _utils.isSuperDir)(subdirOld, resolvePath);
        if (!isInnerFile) {
          continue;
        }

        const newPath = _path2.default.join(subdirNew, _path2.default.relative(subdirOld, resolvePath));
        let newDep = isRela ? _path2.default.normalize(_path2.default.relative(dir, (0, _utils.fsToDep)(newPath))) : (0, _module_resolver.useImplicitPrefix)((0, _utils.fsToDep)(newPath));
        if (isRela && !(0, _utils.isRelative)(newDep)) {
          newDep = `./${newDep}`;
        }
        console.log('update file: %s\n  %s\n->%s\n', file, dep, newDep);
        fileContent = fileContent.substr(0, m.index) + fileContent.substr(m.index).replace(dep, newDep);
        re.lastIndex += newDep.length - dep.length;
      }
    });
    _fs2.default.writeFileSync(file, fileContent);
  });
}

if (isDirectory) {
  fixFilesInSubdir();
  fixFilesOutsideSubdir();
} else {
  fixOnlyFile();
  fixFilesDependingOnOnlyFile();
}
_fs2.default.renameSync(fromFile, toFile);